/*****************************************************************************
 *
 * ethernet_util.c -- MAC address utility and function
 *
 *
 * Copyright (C) Panasonic Corporation
 *
 *****************************************************************************/

/* Configuration */
#define _USE_INT

#include <common.h>
#include <asm/system.h>

#ifdef _USE_INT
//#include <asm/arch/irq.h>
//#include <asm/arch/intctl-regs.h>
#endif

#include <asm/io.h>
#include <asm/arch/cpu-regs.h>
#include <asm/arch/timer-regs.h>

static int eep_get_mac(unsigned char *mac_addr);
static int eep_set_mac(unsigned char *mac_addr);
static int check_busy(unsigned int factor);
static int wait_interrupt(unsigned int factor, unsigned int *status);

/***************************************************/
/* MAC address get & set utility                   */
/***************************************************/
int util_mac_read(unsigned char *mac_buf)
{
	int ret;

	mac_buf[0] = 0xff;
	mac_buf[1] = 0xff;
	mac_buf[2] = 0xff;
	mac_buf[3] = 0xff;
	mac_buf[4] = 0xff;
	mac_buf[5] = 0xff;

	ret = eep_get_mac(mac_buf);

	return ret;
}

int util_mac_write(unsigned char *mac_buf)
{
	int ret;
	
	ret = eep_set_mac(mac_buf);

	return ret;
}


/***************************************************/
/* MAC address get & set function                  */
/***************************************************/
#if 1 /* <--START sharp modified 2015/01/16 */
#define EEPROM_ADDRESS			(0xAC)
#define VALID_MAC_EXIST_SZ      (2)
#define VALID_MAC_CHK_MAGIC_NUM (0x1234) // This magic number indicates if the mac address is valid or not.
#define VALID_MAC_CHK_OFFSET    (0x1b84) // magic number (2Bytes) + Mac address (6Bbytes)
#define MAC_OFFSET              (VALID_MAC_CHK_OFFSET+VALID_MAC_EXIST_SZ)
#else /* original source code as follows */
#define EEPROM_ADDRESS			(0xA0)

#define MAC_OFFSET				(0xFFF9)
#endif/* <--END sharp modified 2015/01/16 */

#define MAC_LENGTH				(6)

#define USE_IIC_CH				(0)       /* IIC CH 0,1,2,3 */

#define IIC_OFFSET				(0x1000 * USE_IIC_CH)

#define IIC_FIFO_SIZE			(8)

#define IIC_BASE				(0x58780000)

#define IIC_FI2CCR				__SYSREG(IIC_BASE + IIC_OFFSET)
#define IIC_FI2CDTTX			__SYSREG(IIC_BASE + 0x04 + IIC_OFFSET)
#define IIC_FI2CDTRX			__SYSREG(IIC_BASE + 0x04 + IIC_OFFSET)
#define IIC_FI2CSLAD			__SYSREG(IIC_BASE + 0x0c + IIC_OFFSET)
#define IIC_FI2CCYC				__SYSREG(IIC_BASE + 0x10 + IIC_OFFSET)
#define IIC_FI2CLCTL			__SYSREG(IIC_BASE + 0x14 + IIC_OFFSET)
#define IIC_FI2CSSUT			__SYSREG(IIC_BASE + 0x18 + IIC_OFFSET)
#define IIC_FI2CDSUT			__SYSREG(IIC_BASE + 0x1c + IIC_OFFSET)
#define IIC_FI2CINT				__SYSREG(IIC_BASE + 0x20 + IIC_OFFSET)
#define IIC_FI2CIE				__SYSREG(IIC_BASE + 0x24 + IIC_OFFSET)
#define IIC_FI2CIC				__SYSREG(IIC_BASE + 0x28 + IIC_OFFSET)
#define IIC_FI2CSR				__SYSREG(IIC_BASE + 0x2c + IIC_OFFSET)
#define IIC_FI2CRST				__SYSREG(IIC_BASE + 0x34 + IIC_OFFSET)
#define IIC_FI2CBM				__SYSREG(IIC_BASE + 0x38 + IIC_OFFSET)
#define IIC_FI2CBRST			__SYSREG(IIC_BASE + 0x50 + IIC_OFFSET)
#define MP_IRQSEL				__SYSREG(0x59820050)

#define IIC_FI2CSR_MD_MASK		((unsigned int)0x00006000)
#define IIC_FI2CSR_DB_MASK		((unsigned int)0x00001000)
#define IIC_FI2CSR_STS_MASK		((unsigned int)0x00000800)
#define IIC_FI2CSR_LRB_MASK		((unsigned int)0x00000400)
#define IIC_FI2CSR_ASS_MASK		((unsigned int)0x00000200)
#define IIC_FI2CSR_BB_MASK		((unsigned int)0x00000100)
#define IIC_FI2CSR_GCA_MASK		((unsigned int)0x00000010)
#define IIC_FI2CSR_RFF_MASK		((unsigned int)0x00000008)
#define IIC_FI2CSR_RNE_MASK		((unsigned int)0x00000004)
#define IIC_FI2CSR_TNF_MASK		((unsigned int)0x00000002)
#define IIC_FI2CSR_TFE_MASK		((unsigned int)0x00000001)

#define IIC_FI2CI_TE_MASK		((unsigned int)0x00000200)
#define IIC_FI2CI_RB_MASK		((unsigned int)0x00000010)
#define IIC_FI2CI_NA_MASK		((unsigned int)0x00000004)
#define IIC_FI2CI_AL_MASK		((unsigned int)0x00000002)

#define IIC_FI2CCR_MST_MASK		((unsigned int)0x00000008)
#define IIC_FI2CCR_STA_MASK		((unsigned int)0x00000004)
#define IIC_FI2CCR_STO_MASK		((unsigned int)0x00000002)
#define IIC_FI2CCR_ACK_MASK		((unsigned int)0x00000001)

#define IIC_FI2CDTTX_COM_MASK	((unsigned int)0x00000100)
#define IIC_FI2CDTTX_DATA_MASK	((unsigned int)0x000000FF)

#define IIC_CLK_INTERNAL		((unsigned int)50000000)
#define IIC_CLK_CYCLE			((unsigned int)(IIC_CLK_INTERNAL / 100000))
#define IIC_CLK_LOW_PERIOD		((unsigned int)(4700 / (10000/IIC_CLK_CYCLE)))

#define IIC_IRQ					(73 + USE_IIC_CH)
#define IIC_GIC_PEND_CLR		__SYSREG(0x5FC20380 + (IIC_IRQ / 32) * 4)
#define IIC_GIC_PEND_MASK		(1 << (IIC_IRQ % 32))

#define IIC_DELAY				1000		/* usec */
#define IIC_TIMEOUT				1000000		/* usec */

#define inl(a)					(*((volatile unsigned long*)(a)))
#define outl(d, a)				(*((volatile unsigned long*)(a)) = d)

#ifdef _USE_INT
#define SET_INT(factor)			{										\
	outl((factor | IIC_FI2CI_NA_MASK | IIC_FI2CI_AL_MASK), IIC_FI2CIE);	\
	outl((factor | IIC_FI2CI_NA_MASK | IIC_FI2CI_AL_MASK), IIC_FI2CIC);	\
}
#else	/* _USE_INT */
#define SET_INT(factor)
#endif	/* _USE_INT */


static int eep_get_mac(unsigned char *mac_addr)
{
	unsigned char *mac_buf = mac_addr;
	unsigned int data;
	unsigned int status = 0;
	int mac_length = 6;
	
	/* select FI2C */
	outl(0xB600007F, MP_IRQSEL);

	/* enable bus */
	outl(0x00000003, IIC_FI2CBRST);

	/* set clock */
	outl(IIC_CLK_CYCLE, IIC_FI2CCYC);
	outl(IIC_CLK_LOW_PERIOD, IIC_FI2CLCTL);

	/* check device busy */
	if(check_busy(IIC_FI2CSR_DB_MASK)){
		return (-1);
	}
	/* write slave address */
	data = (EEPROM_ADDRESS & IIC_FI2CDTTX_DATA_MASK) | IIC_FI2CDTTX_COM_MASK;
	outl(data, IIC_FI2CDTTX);
	
	/* write offset address of MAC address */
	data = (MAC_OFFSET & 0xFF00) >> 8;
	outl(data, IIC_FI2CDTTX);

	data = (MAC_OFFSET & 0x00FF);
	outl(data, IIC_FI2CDTTX);

	/* enable interrupt */
	SET_INT(IIC_FI2CI_TE_MASK);
	
	/* check bus busy */
	if(check_busy(IIC_FI2CSR_BB_MASK)){
		return (-1);
	}
	/* set Master mode and start condition */
	data = IIC_FI2CCR_MST_MASK | IIC_FI2CCR_STA_MASK;
	outl(data, IIC_FI2CCR);
	
	/* wait interrunt */
	if(wait_interrupt(IIC_FI2CI_TE_MASK, &status)){
		printf("Error!! %s(%d)\n", __FUNCTION__, __LINE__);
		return (-1);
	}
	
	/* write slave address, start condition and read mode */
	data = (EEPROM_ADDRESS & IIC_FI2CDTTX_DATA_MASK) | IIC_FI2CDTTX_COM_MASK | 0x00000001;
	outl(data, IIC_FI2CDTTX);
	
	/* enable interrupt */
	SET_INT(IIC_FI2CI_RB_MASK);
		
	/* set Master mode and start condition */
	data = IIC_FI2CCR_MST_MASK | IIC_FI2CCR_STA_MASK;
	outl(data, IIC_FI2CCR);
	
	/* receive MAC address */
	while (mac_length > 0) {
		/* wait interrunt */
		if(wait_interrupt(IIC_FI2CI_RB_MASK, &status)){
			printf("Error!! %s(%d)\n", __FUNCTION__, __LINE__);
			return (-1);
		}
		data = inl(IIC_FI2CDTRX);
		*mac_buf++ = (unsigned char)data;
		mac_length--;
		
		if (mac_length == 0) {
			/* stop condition */
			data = IIC_FI2CCR_MST_MASK | IIC_FI2CCR_STO_MASK;
			outl(data, IIC_FI2CCR);
		}
		else if (mac_length == 1) {
			/* send NACK */
			data = IIC_FI2CCR_MST_MASK | IIC_FI2CCR_ACK_MASK;
			outl(data, IIC_FI2CCR);
		}
		else {
			/* send ACK */
		}
		/* clear interrupt */
		outl(status, IIC_FI2CIC);
	}
	return 0;
}

static int eep_set_mac(unsigned char *mac_addr)
{
#if 1 /* <--START sharp modified 2015/01/16 */
	unsigned char mac_data_buf[8];
	unsigned char *mac_buf = mac_data_buf;
#else /* original source code as follows */
	unsigned char *mac_buf = mac_addr;
#endif/* <--END sharp modified 2015/01/16 */
	unsigned int data;
	unsigned int status = 0;
#if 1 /* <--START sharp modified 2015/01/16 */
	int mac_length = MAC_LENGTH + VALID_MAC_EXIST_SZ;
#else /* original source code as follows */
	int mac_length = MAC_LENGTH;
#endif/* <--END sharp modified 2015/01/16 */
	int len = IIC_FIFO_SIZE;

#if 1 /* <--START sharp modified 2015/01/16 */
	{
		/* magic number */	mac_data_buf[0] = (VALID_MAC_CHK_MAGIC_NUM & 0xFF00) >> 8;
		/* magic number */	mac_data_buf[1] = VALID_MAC_CHK_MAGIC_NUM & 0xff;
		/* mac address  */	mac_data_buf[2] = 0xff;
		/* mac address  */	mac_data_buf[3] = 0xff;
		/* mac address  */	mac_data_buf[4] = 0xff;
		/* mac address  */	mac_data_buf[5] = 0xff;
		/* mac address  */	mac_data_buf[6] = 0xff;
		/* mac address  */	mac_data_buf[7] = 0xff;
		
		memcpy( &mac_data_buf[2], mac_addr, MAC_LENGTH);
	}
#endif/* <--END sharp modified 2015/01/16 */
	/* select FI2C */
	outl(0xB600007F, MP_IRQSEL);
	
	/* enable bus */
	outl(0x00000003, IIC_FI2CBRST);

	/* set clock */
	outl(IIC_CLK_CYCLE, IIC_FI2CCYC);
	outl(IIC_CLK_LOW_PERIOD, IIC_FI2CLCTL);

	/* check device busy */
	if(check_busy(IIC_FI2CSR_DB_MASK)){
		return (-1);
	}
	/* write slave address */
	data = (EEPROM_ADDRESS & IIC_FI2CDTTX_DATA_MASK) | IIC_FI2CDTTX_COM_MASK;
	outl(data, IIC_FI2CDTTX);
	len--;
	
	/* write offset address of MAC address */
#if 1 /* <--START sharp modified 2015/01/16 */
	data = (VALID_MAC_CHK_OFFSET & 0xFF00) >> 8;
#else /* original source code as follows */
	data = (MAC_OFFSET & 0xFF00) >> 8;
#endif/* <--END sharp modified 2015/01/16 */
	outl(data, IIC_FI2CDTTX);
	len--;

#if 1 /* <--START sharp modified 2015/01/16 */
	data = (VALID_MAC_CHK_OFFSET & 0x00FF);
#else  /* original source code as follows */
	data = (MAC_OFFSET & 0x00FF);
#endif/* <--END sharp modified 2015/01/16 */
	outl(data, IIC_FI2CDTTX);
	len--;
	
	/* write MAC address (fist write) */
	while (len > 0) {
		data = *mac_buf++ & IIC_FI2CDTTX_DATA_MASK;
		outl(data, IIC_FI2CDTTX);
		mac_length--;
		len--;
	}

	/* enable interrupt */
	SET_INT(IIC_FI2CI_TE_MASK);

	/* check bus busy */
	if(check_busy(IIC_FI2CSR_BB_MASK)){
		return (-1);
	}
	/* set Master mode and start condition */
	data = IIC_FI2CCR_MST_MASK | IIC_FI2CCR_STA_MASK;
	outl(data, IIC_FI2CCR);
	
	/* wait interrunt */
	if(wait_interrupt(IIC_FI2CI_TE_MASK, &status)){
		printf("Error!! %s(%d)\n", __FUNCTION__, __LINE__);
		return (-1);
	}

	/* write MAC address (second write) */
	while (mac_length > 0) {
		data = *mac_buf++ & IIC_FI2CDTTX_DATA_MASK;
		outl(data, IIC_FI2CDTTX);
		mac_length--;
	}
	/* wait interrunt */
	if(wait_interrupt(IIC_FI2CI_TE_MASK, &status)){
		printf("Error!! %s(%d)\n", __FUNCTION__, __LINE__);
		return (-1);
	}
	/* stop condition */
	data = IIC_FI2CCR_MST_MASK | IIC_FI2CCR_STO_MASK;
	outl(data, IIC_FI2CCR);

	return 0;
}

static int check_busy(unsigned int factor){
	int i;
	
	for(i = 0 ; i < (IIC_TIMEOUT/IIC_DELAY) ; i++){
		udelay(IIC_DELAY);
		if(0 == (inl(IIC_FI2CSR) & factor)){
			return 0;
		}
	}
	printf("EEPROM device or I2C bus busy!! : 0x%x!!\n", factor);
	return (-1);
}


static int wait_interrupt(unsigned int factor, unsigned int *status){
	int ret = 0;
#ifdef _USE_INT
	int i;
	unsigned int status_int, status_cnt;
	
	for(i = 0 ; i < (IIC_TIMEOUT/IIC_DELAY) ; i++){
		udelay(IIC_DELAY);
		/* read interrupt status */
		status_int = inl(IIC_FI2CINT);
		if(status_int & IIC_FI2CI_TE_MASK){
			break;
		}
		else if(status_int & IIC_FI2CI_RB_MASK){
			/* return without clear interrupt */
			*status = status_int;
			return 0;
		}
		else if(status_int & IIC_FI2CI_NA_MASK){
			/* stop condition */
			outl((IIC_FI2CCR_MST_MASK | IIC_FI2CCR_STO_MASK), IIC_FI2CCR);
			printf("No reply from EEPROM!!\n");
			ret = (-1);
			break;
		}
		else if(status_int & IIC_FI2CI_AL_MASK){
			printf("Lost arbitrarion to EEPROM!!\n");
			ret = (-1);
			break;
		}
#if 0
		else if(status_int){
			printf("Unknown interrupt from EEPROM!! : 0x%x\n", status_int);
			ret = (-1);
			break;
		}
#endif
	}
	if((IIC_TIMEOUT/IIC_DELAY) == i){
		status_cnt = inl(IIC_FI2CSR);
		printf("Timeout wait for interrunt!! INT=0x%x SR=0x%x\n", status_int, status_cnt);
		ret = (-1);
	}
	/* clear interrupt */
	outl(status_int, IIC_FI2CIC);
#else	/* _USE_INT */
	udelay(IIC_DELAY);
#endif	/* _USE_INT */
	return ret;
}
