/*
 * Copyright (C) 2010 Panasonic Corporation
 * All Rights Reserved.
 *
 * SPDX-License-Identifier:	GPL-2.0+
 */

#include <common.h>
#include <command.h>
#include <asm/arch/serial-regs.h>
#include <asm/arch/timer-regs.h>
#include <serial.h>
#include <linux/compiler.h>

DECLARE_GLOBAL_DATA_PTR;

/*
 * Local function prototypes
 */

/*
 * Minimal global serial functions needed to use one of the SCI modules.
 */

#ifdef CONFIG_UNIPHIER_SERIAL_HIGH_SPEED		/* High Speed serial */
#define SCTXB		SC4TDR
#define SCRXB		SC4RDR
#define SCWAIT_TBF	while (!(*SC4LSR & SC4LSR_TDRE)) continue;
#define SCWAIT_RBF	while (!(*SC4LSR & SC4LSR_RDR)) continue;
#define SCCHECK_RBF	(*SC4LSR & SC4LSR_RDR)

#else	/* Low Speed serial */
#error "No Support Low Speed serial"
#endif	/* CONFIG_UNIPHIER_SERIAL_HIGH_SPEED */

#ifdef CONFIG_SC_UART
#define support_card_base	gd->arch.support_card_base

static void serial_support_card_init(void)
{
	unsigned long quot;
	/* refer to uart_get_divisor in serial_core.c for this setting */
	quot=(UARTCLK+8*CONFIG_BAUDRATE)/ (16*CONFIG_BAUDRATE);

	/* 1. clear the reset of support card
	 * clear reset in board_init. refer to
	 * u-boot/board/panasonic/MN2WS0230_ref/MN2WS0230_ref.c
	 */

	/* 2. enable FIFO */
	*(volatile unsigned char*)(CONFIG_UNIPHIER_SERIAL_BASE+UNIPHIER_UART_FCR)=0xC1;

	/* 3. Set DLAB */
	*(volatile unsigned char*)(CONFIG_UNIPHIER_SERIAL_BASE+UNIPHIER_UART_LCR)=UNIPHIER_UART_LCR_WLEN8|0x80;

	/* 4. set LS of divisor */
	//*(volatile unsigned char*)(CONFIG_UNIPHIER_SERIAL_BASE+UNIPHIER_UART_DLL)=0x07;
	*(volatile unsigned char*)(CONFIG_UNIPHIER_SERIAL_BASE+UNIPHIER_UART_DLL)=quot & 0xff;

	/* 5. set MS of divisor */
	//*(volatile unsigned char*)(CONFIG_UNIPHIER_SERIAL_BASE+UNIPHIER_UART_DLM)=0x00;
	*(volatile unsigned char*)(CONFIG_UNIPHIER_SERIAL_BASE+UNIPHIER_UART_DLM)=quot>>8;

	/* 6. reset DLAB */
	*(volatile unsigned char*)(CONFIG_UNIPHIER_SERIAL_BASE+UNIPHIER_UART_LCR)=UNIPHIER_UART_LCR_WLEN8;

	/* 7. disable interrupt */
	*(volatile unsigned char*)(CONFIG_UNIPHIER_SERIAL_BASE+UNIPHIER_UART_IER)=0x0;
}
#endif /*CONFIG_SC_UART*/

static int uniphier_serial_init(void)
{
#ifdef CONFIG_SC_UART
	serial_support_card_init();
#else /* CONFIG_SC_UART */

	/* Set Serial Port */
#ifdef CONFIG_UNIPHIER_SERIAL_HIGH_SPEED
	*SC4FCR = SC4FCR_FEN;
	*SC4LCR = SC4LCR_CHL;
#endif	/* CONFIG_UNIPHIER_SERIAL_HIGH_SPEED */

	/* Set baudrate */
	*SC4DLR = UARTCLK / (gd->baudrate) / 16;
#endif /* CONFIG_SC_UART */

	return 0;
}

static void uniphier_serial_putc(const char c)
{
    if (c == '\n') {
		SCWAIT_TBF
		*SCTXB = '\r';
	}
	SCWAIT_TBF
    *SCTXB = c;
}

static int uniphier_serial_getc(void)
{
	int c = '\0';

	SCWAIT_RBF
	c = *SCRXB;
	if(c == '\r')
		c = '\n';
	return c;
}

static int uniphier_serial_tstc(void)
{
	return SCCHECK_RBF;
}

static void uniphier_serial_setbrg(void)
{
	uniphier_serial_init();
	return;
}

static struct serial_device uniphier_serial_drv = {
	.name   = "uniphier_serial",
	.start  = uniphier_serial_init,
	.stop   = NULL,
	.setbrg = uniphier_serial_setbrg,
	.putc   = uniphier_serial_putc,
	.puts   = default_serial_puts,
	.getc   = uniphier_serial_getc,
	.tstc   = uniphier_serial_tstc,
};

void uniphier_serial_initialize(void)
{
	serial_register(&uniphier_serial_drv);
}

__weak struct serial_device *default_serial_console(void)
{
	return &uniphier_serial_drv;
}
