/*
 * Copyright (C) 2010 Panasonic Corporation
 * All Rights Reserved.
 *
 * SPDX-License-Identifier:	GPL-2.0+
 */

#ifndef _ASM_ARCH_SERIAL_REGS_H
#define _ASM_ARCH_SERIAL_REGS_H

#include <asm/arch/cpu-regs.h>

#ifdef CONFIG_SC_UART

#define UARTCLK			12288000

#define UNIPHIER_UART_REG_SHIFT	1

#define UNIPHIER_UART_RX	(0x0<<UNIPHIER_UART_REG_SHIFT)
#define UNIPHIER_UART_TX	(0x0<<UNIPHIER_UART_REG_SHIFT)
#define UNIPHIER_UART_IER	(0x1<<UNIPHIER_UART_REG_SHIFT)
#define UNIPHIER_UART_IIR	(0x2<<UNIPHIER_UART_REG_SHIFT)
#define UNIPHIER_UART_FCR	(0x2<<UNIPHIER_UART_REG_SHIFT)
#define UNIPHIER_UART_LCR	(0x3<<UNIPHIER_UART_REG_SHIFT)
#define UNIPHIER_UART_MCR	(0x4<<UNIPHIER_UART_REG_SHIFT)
#define UNIPHIER_UART_LSR	(0x5<<UNIPHIER_UART_REG_SHIFT)
#define UNIPHIER_UART_MSR	(0x6<<UNIPHIER_UART_REG_SHIFT)
#define UNIPHIER_UART_DLL	(0x0<<UNIPHIER_UART_REG_SHIFT)
#define UNIPHIER_UART_DLM	(0x1<<UNIPHIER_UART_REG_SHIFT)

#define UNIPHIER_UART_LCR_WLEN8	(0x03)

#else	/* CONFIG_SC_UART */

#if defined(__MN2WS0220_ref_H) || defined(__MN2WS0251_ref_H)
#define UARTCLK			36864000
#elif defined(__MN2WS0230_ref_H)
#define UARTCLK			73728000
#elif defined(__MN2WS0270_ref_H)
#define UARTCLK			80000000
#elif defined(__MN2WS0310_ref_H)
#define UARTCLK			88889000
#endif

#define UNIPHIER_UART_REG_SHIFT	0

#define UNIPHIER_UART_RX	0x00
#define UNIPHIER_UART_TX	0x00
#define UNIPHIER_UART_IER	0x04
#define UNIPHIER_UART_IIR	0x08
#define UNIPHIER_UART_FCR	0x0c
#define UNIPHIER_UART_LCR	0x10
#define UNIPHIER_UART_MCR	0x10
#define UNIPHIER_UART_LSR	0x14
#define UNIPHIER_UART_MSR	0x18
#define UNIPHIER_UART_DLL	0x24

#define UNIPHIER_UART_LCR_WLEN8	(UART_LCR_WLEN8<<8)
#endif /* CONFIG_SC_UART */

/* serial port 4 */
#define	SC4RDR				__SYSREGB(CONFIG_UNIPHIER_SERIAL_BASE)		/* receive buffer reg */
#define	SC4TDR				__SYSREGB(CONFIG_UNIPHIER_SERIAL_BASE)		/* transmit buffer reg */

#define	SC4IER				__SYSREGB(CONFIG_UNIPHIER_SERIAL_BASE + UNIPHIER_UART_IER)		/* interrupt control reg */

#define	SC4IIR				__SYSREGB(CONFIG_UNIPHIER_SERIAL_BASE + UNIPHIER_UART_IIR)		/* interrupt indication reg */

#define	SC4FCR				__SYSREGB(CONFIG_UNIPHIER_SERIAL_BASE + UNIPHIER_UART_FCR)		/* FIFO control reg */
#define	SC4FCR_TFR			0x04											/* transmit FIFO reset */
#define	SC4FCR_RFR			0x02											/* receive FIFO reset */
#define	SC4FCR_FEN			0x01											/* FIFO enable */

#define	SC4LCR				__SYSREGH(CONFIG_UNIPHIER_SERIAL_BASE + UNIPHIER_UART_LCR)		/* line control reg */
#define	SC4LCR_BRC			0x4000											/* transmit brake */
#define	SC4LCR_STP			0x2000											/* select parity */
#define	SC4LCR_EPS			0x1000											/* enable parity */
#define	SC4LCR_PE			0x0800											/* stop bit length */
#define	SC4LCR_CHL			0x0300											/* character length */

#define	SC4MCR				__SYSREGH(CONFIG_UNIPHIER_SERIAL_BASE + UNIPHIER_UART_MCR)		/* modem control reg */
#define	SC4MCR_AFE			0x0020											/* auto flow control */
#define	SC4MCR_LPT			0x0010											/* select clock source */
#define	SC4MCR_RTS			0x0002											/* XRTS output level */

#define	SC4LSR				__SYSREGB(CONFIG_UNIPHIER_SERIAL_BASE + UNIPHIER_UART_LSR)		/* line status reg */
#define	SC4LSR_FFE			0x80											/* FIFO error */
#define	SC4LSR_TE			0x40											/* transmit status */
#define	SC4LSR_TDRE			0x20											/* buffer status */
#define	SC4LSR_BRI			0x10											/* brake interrupt */
#define	SC4LSR_FME			0x08											/* frame error */
#define	SC4LSR_PTE			0x04											/* parity error */
#define	SC4LSR_OVE			0x02											/* overrun error */
#define	SC4LSR_RDR			0x01											/* receive status */

#define	SC4MSR				__SYSREGB(CONFIG_UNIPHIER_SERIAL_BASE + UNIPHIER_UART_MSR)		/* modem status reg */

#define SC4DLR				__SYSREGH(CONFIG_UNIPHIER_SERIAL_BASE + UNIPHIER_UART_DLL)		/* baudrate set reg */
#ifdef CONFIG_SC_UART
#define	SC4DLM				__SYSREGB(CONFIG_UNIPHIER_SERIAL_BASE + UNIPHIER_UART_DLM)
#endif


#endif /* _ASM_ARCH_SERIAL_REGS_H */
