/*
 * Copyright (C) 2014 Panasonic Corporation
 * All Rights Reserved.
 *
 * SPDX-License-Identifier:	GPL-2.0
 */

#include <common.h>
#include <mmc.h>
#include <malloc.h>
#include <asm/io.h>
#include <asm/errno.h>
#include <asm/byteorder.h>
#include <asm/arch/cacheflush.h>
#include "uniphier_mmc.h"

#define DEBUG	0

#if DEBUG
#define DBG_PRT_REG(fmt, args...) printf(fmt, ##args)
#else
#define DBG_PRT_REG(fmt, args...)
#endif

#if DEBUG
#define DBG_PRT_GEN(fmt, args...) printf(fmt, ##args)
#else
#define DBG_PRT_GEN(fmt, args...)
#endif
#if DEBUG
#define DBG_PRT_FUNC(fmt, args...)	printf("%s: " fmt, __FUNCTION__, ## args)
#else
#define DBG_PRT_FUNC(fmt, args...)
#endif


static u32 read_32x1(u32 addr)
{
 	u32 ret;

	DBG_PRT_REG("R , addr:0x%08x\n", addr);

	ret = *(volatile u32 *)addr;

	DBG_PRT_REG("R , ret value:0x%08x\n", ret);

	return ret;
}

static u32 maskread_32x1(u32 addr, u32 mask)
{
	u32 ret;

	DBG_PRT_REG("MR, addr:0x%08x, mask:0x%08x\n", addr, mask);

	ret = read_32x1(addr) & mask;
	DBG_PRT_REG("MR, ret value:0x%08x\n", ret);

	return ret;
}

static u32 write_32x1(u32 addr, u32 data)
{
	DBG_PRT_REG("W , addr:0x%08x, data:0x%08x\n", addr, data);

	*(volatile u32 *)addr = data;

	DBG_PRT_REG("W , ret value:0x%08x\n", data);

	return data;
}

static u32 maskwrite_32x1(u32 addr, u32 mask, u32 data)
{
	u32 value;
	
	DBG_PRT_REG("MW, addr:0x%08x, mask:0x%08x, data:0x%08x\n", addr, mask, data);

	value = (read_32x1(addr) & ~(mask)) | (data & mask);

	write_32x1(addr, value);

	DBG_PRT_REG("MW, ret value:0x%08x\n", value);

	return value;
}

#if defined(CONFIG_MN2WS0220_ref) || defined(CONFIG_MN2WS0230_ref)
static int mio_dma_status(void)
{
	return 0;
}

static void mio_dma_set_adrmode(struct uniphier_mmc *host, int rw)
{
	int dma_ch = host->dma_ch;
	u32 dmac_base = DMAC_BASEADDR;

	u32 ChSrcAmode = 0;
	u32 ChDstAmode = 0;

	/* Ring buffer setting */
	if(rw == MMC_DATA_READ){
		/* MMC -> DDR */
		ChSrcAmode |= DMAC_CDMIO_CHAMODE_ENDIAN0;	/* Little-Endian */
		ChSrcAmode |= DMAC_CDMIO_CHAMODE_AUPDT2;	/* Fixed */
		ChSrcAmode |= DMAC_CDMIO_CHAMODE_TYPE0;		/* Normal memory */

		ChDstAmode |= DMAC_CDMIO_CHAMODE_ENDIAN0;	/* Little-Endian */
		ChDstAmode |= DMAC_CDMIO_CHAMODE_AUPDT0;	/* Increment */
		ChDstAmode |= DMAC_CDMIO_CHAMODE_TYPE0;		/* Normal memory */
	}else{
		/* DDR -> MMC */
		ChSrcAmode |= DMAC_CDMIO_CHAMODE_ENDIAN0;	/* Little-Endian */
		ChSrcAmode |= DMAC_CDMIO_CHAMODE_AUPDT0;	/* Increment */
		ChSrcAmode |= DMAC_CDMIO_CHAMODE_TYPE0;		/* Normal memory */

		ChDstAmode |= DMAC_CDMIO_CHAMODE_ENDIAN0;	/* Little-Endian */
		ChDstAmode |= DMAC_CDMIO_CHAMODE_AUPDT2;	/* Fixed */
		ChDstAmode |= DMAC_CDMIO_CHAMODE_TYPE0;		/* Normal memory */
	}

	ChSrcAmode |= (dma_ch << 8);
	ChDstAmode |= (dma_ch << 8);

	write_32x1(dmac_base + REG_DMAC_CDMIO_CHxSRCAMODE(dma_ch), ChSrcAmode);
	write_32x1(dmac_base + REG_DMAC_CDMIO_CHxDSTAMODE(dma_ch), ChDstAmode);
}

static void mio_dma_set_transfer(struct uniphier_mmc *host, int rw, u32 Addr, u32 Count)
{
	int dma_ch = host->dma_ch;
	u32 dmac_base = DMAC_BASEADDR;

	u32 CountByte;

	CountByte = Count;

	write_32x1(dmac_base + REG_DMAC_CDMIO_CHxIE(dma_ch), DMAC_CDMIO_CH_ALLCLR);
	maskwrite_32x1(dmac_base + REG_DMAC_CDMIO_CHxIR(dma_ch),
		DMAC_CDMIO_CH_CH1, DMAC_CDMIO_CH_CH1);

	/* IRQ enable */
	maskwrite_32x1(dmac_base + REG_DMAC_CDMIO_CHxIE(dma_ch),
		DMAC_CDMIO_CH_CH1, DMAC_CDMIO_CH_CH1);

	if(rw == MMC_DATA_READ){
		/* MMC -> DDR */
		write_32x1(dmac_base + REG_DMAC_CDMIO_CHxSRCSTRTADRS(dma_ch), 0);
		write_32x1(dmac_base + REG_DMAC_CDMIO_CHxDSTSTRTADRS(dma_ch), Addr);
	}else{
		/* DDR -> eMMC */
		write_32x1(dmac_base + REG_DMAC_CDMIO_CHxSRCSTRTADRS(dma_ch), Addr);
		write_32x1(dmac_base + REG_DMAC_CDMIO_CHxDSTSTRTADRS(dma_ch), 0);
	}

	write_32x1(dmac_base + REG_DMAC_CDMIO_CHxSIZE(dma_ch), CountByte);

	return;
}

static void mio_dma_start(struct uniphier_mmc *host)
{
	u32 Strt0 = 0;

	Strt0 |= DMAC_CDMIO_STRT0_REQ(host->dma_ch);
	Strt0 &= ~DMAC_CDMIO_STRT0_STOP;

	/* DMA start */
	write_32x1(DMAC_BASEADDR + REG_DMAC_CDMIO_STRT0, Strt0);

	return;
}

static int mio_dma_checkdmaend(struct uniphier_mmc *host)
{
	int dma_ch = host->dma_ch;
	u32 dmac_base = DMAC_BASEADDR;

	int   count = 0;
	int  ret = 1;
	u32 value;

	for(count = 0;count < 1000;count++){
		value = read_32x1(dmac_base + REG_DMAC_CDMIO_CHxIR(dma_ch));
		if(value & DMAC_CDMIO_CH_CH1){
			ret = 0;
			maskwrite_32x1(dmac_base + REG_DMAC_CDMIO_CHxIR(dma_ch),
				DMAC_CDMIO_CH_CH1, DMAC_CDMIO_CH_CH1);
			break;
		}
	}

	return ret;
}
#endif /* CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref */

static int mmc_dma_status(void)
{
#if defined(CONFIG_MN2WS0220_ref) || defined(CONFIG_MN2WS0230_ref)
	return mio_dma_status();
#else /* CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref */
	/* Skip status check */
	return 0;
#endif /* CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref */
}

static void mmc_dma_set_mode(struct uniphier_mmc *host, int rw)
{
#if defined(CONFIG_MN2WS0220_ref) || defined(CONFIG_MN2WS0230_ref)
	mio_dma_set_adrmode(host, rw);
#else /* CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref */
	write_32x1(host->reg_sdip_base + DM_CM_DTRAN_MODE,
		   DM_CM_DTRAN_MODE_BUS_WIDTH_64 |
		   DM_CM_DTRAN_MODE_ADDR_MODE_INCREMENT |
		   ((rw == MMC_DATA_READ) ? DM_CM_DTRAN_MODE_CH_NUM_FROM_MMC
					  : DM_CM_DTRAN_MODE_CH_NUM_TO_MMC)
		   );
#endif /* CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref */
}

static void mmc_dma_set_transfer(struct uniphier_mmc *host, int rw, u32 Addr,
				 u32 Count)
{
#if defined(CONFIG_MN2WS0220_ref) || defined(CONFIG_MN2WS0230_ref)
	mio_dma_set_transfer(host, rw, Addr, Count);
#else /* CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref */
	write_32x1(host->reg_sdip_base + DM_DTRAN_ADDR,
		   Addr & DM_CM_DTRAN_ADDR_DADDR);
#endif /* CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref */
}

static void mmc_dma_start(struct uniphier_mmc *host, int rw)
{
#if defined(CONFIG_MN2WS0220_ref) || defined(CONFIG_MN2WS0230_ref)
	mio_dma_start(host);
#else /* CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref */
	u32 bit;

	/* Clear interrupt register */
	bit = (rw == MMC_DATA_READ) ? DM_CM_INFO1_DTARNEND_FROM_MMC
				    : DM_CM_INFO1_DTARNEND_TO_MMC;
	maskwrite_32x1(host->reg_sdip_base + DM_CM_INFO1, bit, ~bit);

	/* Clear interrupt mask register */
	bit = (rw == MMC_DATA_READ) ? DM_CM_INFO1_MASK_DTARNEND_MASK_FROM_MMC
				    : DM_CM_INFO1_MASK_DTARNEND_MASK_TO_MMC;
	maskwrite_32x1(host->reg_sdip_base + DM_CM_INFO1_MASK, bit, ~bit);

	/* Clear error interrupt register */
	bit = (rw == MMC_DATA_READ) ? DM_CM_INFO2_DTARNERR_FROM_MMC
				    : DM_CM_INFO2_DTARNERR_TO_MMC;
	maskwrite_32x1(host->reg_sdip_base + DM_CM_INFO2, bit, ~bit);

	/* Clear error interrupt mask register */
	bit = (rw == MMC_DATA_READ) ? DM_CM_INFO2_MASK_DTARNERR_FROM_MMC
				    : DM_CM_INFO2_MASK_DTARNERR_TO_MMC;
	maskwrite_32x1(host->reg_sdip_base + DM_CM_INFO2_MASK, bit, ~bit);

	/* Start DMA */
	write_32x1(host->reg_sdip_base + DM_CM_DTRAN_CTRL,
		   DM_CM_DTRAN_CTRL_DM_SEL_MMC |
		   DM_CM_DTRAN_CTRL_DM_START);
#endif /* CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref */
}

static int mmc_dma_checkdmaend(struct uniphier_mmc *host, int rw)
{
#if defined(CONFIG_MN2WS0220_ref) || defined(CONFIG_MN2WS0230_ref)
	return mio_dma_checkdmaend(host);
#else /* CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref */
	u32 bit;

	/* Wait until interrupt occurs */
	bit = (rw == MMC_DATA_READ) ? DM_CM_INFO1_DTARNEND_FROM_MMC
				    : DM_CM_INFO1_DTARNEND_TO_MMC;
	while (!(read_32x1(host->reg_sdip_base + DM_CM_INFO1) & bit));

	/* Error check */
	bit = (rw == MMC_DATA_READ) ? DM_CM_INFO2_DTARNERR_FROM_MMC
				    : DM_CM_INFO2_DTARNERR_TO_MMC;
	if (read_32x1(host->reg_sdip_base + DM_CM_INFO2) & bit) {
		return 1;
	}

	return 0;
#endif /* CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref */
}

static int uniphier_mmc_wait_buffer_read_enable(struct mmc *mmc)
{
	uniphier_mmc_t *host = (uniphier_mmc_t *)mmc->priv;
	u32 sdip_base = host->reg_sdip_base;

	u32 sd_info2;

	/* Check SD_INFO2 bit8 BRE */
	while(1){
		sd_info2 = maskread_32x1(sdip_base + REG_SD_INFO2,
				 SD_INFO2_ERROR_ALL | SD_INFO2_INFO_BRE);

		if(sd_info2 & SD_INFO2_ERROR_ALL){
			while(1);
			return 1;
		}else if(sd_info2 & SD_INFO2_INFO_BRE){
			break;
		}
	}

	/* Clear SD_INFO2 bit8 BRE */
	write_32x1(sdip_base + REG_SD_INFO2, ~SD_INFO2_INFO_BRE);

	return 0;
}

static int uniphier_mmc_read_buf(struct mmc *mmc, u32 size, u16 *buf)
{
	uniphier_mmc_t *host = (uniphier_mmc_t *)mmc->priv;

	int ret = 0;
	int i;
	u32 len;
	u32 tmp_data;

	while (size > 0) {
		if (size >= mmc->read_bl_len) {
			len = mmc->read_bl_len;
			size -= mmc->read_bl_len;
		} else {
			len = size;
			size = 0;
		}

		/* Check SD_INFO2 bit8 BRE */
		if(uniphier_mmc_wait_buffer_read_enable(mmc)){
			ret = 1;
			goto end;
		}
		for (i = 0; i < len; i += 4) {
			tmp_data = read_32x1(host->reg_sdip_base + REG_SD_BUF);
			*buf++ = tmp_data & 0xFFFF;
			*buf++ = (tmp_data >> 16) & 0xFFFF;
		}
	}


end:
	return ret;
}

static int uniphier_mmc_wait_buffer_write_enable(struct mmc *mmc)
{
	uniphier_mmc_t *host = (uniphier_mmc_t *)mmc->priv;
	u32 sdip_base = host->reg_sdip_base;

	u32 sd_info2;

	/* Check SD_INFO2 bit9 BWE */
	while(1){
		sd_info2 = maskread_32x1(sdip_base + REG_SD_INFO2,
				SD_INFO2_ERROR_ALL | SD_INFO2_INFO_BWE);

		if(sd_info2 & SD_INFO2_ERROR_ALL){
			while(1);
			return 1;
		}else if(sd_info2 & SD_INFO2_INFO_BWE){
			break;
		}
	}

	/* Clear SD_INFO2 bit9 BWE */
	write_32x1(sdip_base + REG_SD_INFO2, ~SD_INFO2_INFO_BWE);

	return 0;
}

static int uniphier_mmc_write_buf(struct mmc *mmc, u32 size, const u16 *buf)
{
	uniphier_mmc_t *host = (uniphier_mmc_t *)mmc->priv;

	int ret = 0;
	int i;
	u32 len;
	u32 tmp_data;

	while (size > 0) {
		if (size >= mmc->read_bl_len) {
			len = mmc->read_bl_len;
			size -= mmc->read_bl_len;
		} else {
			len = size;
			size = 0;
		}

		/* Check SD_INFO2 bit9 BWE */
		if(uniphier_mmc_wait_buffer_write_enable(mmc)){
			ret = 1;
			goto end;
		}
		for (i = 0; i < len; i += 4) {
			tmp_data = (*buf++) & 0xFFFF;
			tmp_data |= ((*buf++) & 0xFFFF) << 16;
			write_32x1(host->reg_sdip_base + REG_SD_BUF, tmp_data);
		}
	}

end:
	return ret;
}

static int uniphier_mmc_data_end(struct mmc *mmc)
{
	uniphier_mmc_t *host = (uniphier_mmc_t *)mmc->priv;
	u32 sdip_base = host->reg_sdip_base;

	u32 sd_info1;
	u32 sd_info2;

	/* Check SD_INFO1 bit2 */
	while(1){
		sd_info1 = maskread_32x1(sdip_base + REG_SD_INFO1, SD_INFO1_INFO_RWA_CMP);
		sd_info2 = maskread_32x1(sdip_base + REG_SD_INFO2, SD_INFO2_ERROR_ALL);

		if(sd_info2 != 0x0000){
			while(1);
			return 1;
		}else if(sd_info1 == SD_INFO1_INFO_RWA_CMP){
			break;
		}
	}

	/* Clear SD_INFO1 bit2 */
	write_32x1(sdip_base + REG_SD_INFO1, ~SD_INFO1_INFO_RWA_CMP);

	return 0;
}

static int uniphier_send_command_main(struct mmc *mmc, struct mmc_cmd *cmd, struct mmc_data *data)
{
	uniphier_mmc_t *host = (uniphier_mmc_t *)mmc->priv;
	u32 sdip_base = host->reg_sdip_base;

	u32 sd_info1;
	u32 sd_info2;
	int ret = 0;
	u32 command = 0;
	int dma = 0;

#ifndef CONFIG_MMC_UNIPHIER_DMA_DISABLE
	/* Enable DMA only read data or write data */
	if (/*cmd->cmdidx == MMC_CMD_READ_SINGLE_BLOCK   ||*/
	    cmd->cmdidx == MMC_CMD_READ_MULTIPLE_BLOCK ||
	    /*cmd->cmdidx == MMC_CMD_WRITE_SINGLE_BLOCK  ||*/
	    cmd->cmdidx == MMC_CMD_WRITE_MULTIPLE_BLOCK) {
		dma = 1;
	}
#endif /* !CONFIG_MMC_UNIPHIER_DMA_DISABLE */

	/* Check Command Busy Status */
	while (maskread_32x1(sdip_base + REG_SD_INFO2, SD_INFO2_INFO_CBSY) == SD_INFO2_INFO_CBSY);

	/* Clear All IRQ */
	write_32x1(sdip_base + REG_SD_INFO1, 0x0000);
	write_32x1(sdip_base + REG_SD_INFO2, 0x0000);

	if(data){
		command |= SD_CMD_DAT_ADTC;
		write_32x1(sdip_base + REG_SD_SIZE, data->blocksize);

		if(dma){
			/* MMC buffer DMA transfer set(enable) */
			maskwrite_32x1(sdip_base + REG_CC_EXT_MODE, 0x0002, 0x0002);
		}else{
			/* MMC buffer DMA transfer set(disable) */
			maskwrite_32x1(sdip_base + REG_CC_EXT_MODE, 0x0002, 0x0000);
		}

		if(data->blocks > 1){
			/* Enables SD_SECCNT */
			maskwrite_32x1(sdip_base + REG_SD_STOP, 0x0100, 0x0100);
			/* Set SD_SECCNT */
			write_32x1(sdip_base + REG_SD_SECCNT, data->blocks);
			command |= SD_CMD_BK_MULT;
		}else{
			/* Disables SD_SECCNT */
			maskwrite_32x1(sdip_base + REG_SD_STOP, 0x0100, 0x0000);
			command |= SD_CMD_BK_SINGLE;
		}

		if (data->flags == MMC_DATA_WRITE) {
			command |= SD_CMD_RW_WRITE;
		} else {
			command |= SD_CMD_RW_READ;
		}
	}

	write_32x1(sdip_base + REG_SD_ARG0, cmd->cmdarg);

	switch (cmd->resp_type) {
		case MMC_RSP_R1: /* = MMC_RSP_R5 = MMC_RSP_R6 = MMC_RSP_R7 */
			command |= SD_CMD_RESP_R1;
			break;
		case MMC_RSP_R1b:
			if ((cmd->cmdidx == MMC_CMD_READ_SINGLE_BLOCK) ||
			    (cmd->cmdidx == MMC_CMD_STOP_TRANSMISSION)) {
				command |= SD_CMD_RESP_NORESP;
			} else {
				command |= SD_CMD_RESP_R1B;
			}
			break;
		case MMC_RSP_R2:
			command |= SD_CMD_RESP_R2;
			break;
		case MMC_RSP_R3: /* = MMC_RSP_R4 */
			command |= SD_CMD_RESP_R3;
			break;
		case MMC_RSP_NONE:
		default:
			command |= SD_CMD_RESP_NORESP;
			break;
	}

	if (host->appcmd_flg) {
		command |= SD_CMD_APP_ACMD;
		host->appcmd_flg = 0;
	} else {
		command |= SD_CMD_APP_NORMAL;
		if (cmd->cmdidx == MMC_CMD_APP_CMD) {
			host->appcmd_flg = 1;
		}
	}

	command |= cmd->cmdidx;

	if (data) {
		DBG_PRT_GEN("command: 0x%08x arg: 0x%08x "
			"src/dest: 0x%08x blocknum: 0x%08x\n",
			command, cmd->cmdarg, (u32)data->dest, data->blocks);
	} else {
		DBG_PRT_GEN("command: 0x%08x arg: 0x%08x\n",
			command, cmd->cmdarg);
	}

	write_32x1(sdip_base + REG_SD_CMD,  command);

	/* Wait Command response */
	while(1){
		sd_info1 = maskread_32x1(sdip_base + REG_SD_INFO1, SD_INFO1_INFO_CMD_CMP);
		sd_info2 = maskread_32x1(sdip_base + REG_SD_INFO2, SD_INFO2_ERROR_ALL);

		if(sd_info2 != 0x0000){
			DBG_PRT_GEN("Wait command response error! 0x%x\n", sd_info2);
			return sd_info2;
		}else if(sd_info1 == SD_INFO1_INFO_CMD_CMP){
			break;
		}
	}

	/* Clear Command response */
	write_32x1(sdip_base + REG_SD_INFO1, ~SD_INFO1_INFO_CMD_CMP);

	/* Get Response */
	if (cmd->resp_type & MMC_RSP_PRESENT) {
		if (cmd->resp_type & MMC_RSP_136) {
			u32 rsp01, rsp23, rsp45, rsp67;

			rsp01 = read_32x1(sdip_base + REG_SD_RSP0);
			rsp23 = read_32x1(sdip_base + REG_SD_RSP2);
			rsp45 = read_32x1(sdip_base + REG_SD_RSP4);
			rsp67 = read_32x1(sdip_base + REG_SD_RSP6);
			cmd->response[0] = rsp67 << 8 | (rsp45 & 0xff000000) >> 24;
			cmd->response[1] = rsp45 << 8 | (rsp23 & 0xff000000) >> 24;
			cmd->response[2] = rsp23 << 8 | (rsp01 & 0xff000000) >> 24;
			cmd->response[3] = rsp01 << 8;
			DBG_PRT_GEN("resp    : %08x %08x %08x %08x\n",
				cmd->response[0], cmd->response[1],
				cmd->response[2], cmd->response[3]);
		} else {
			cmd->response[0] = read_32x1(sdip_base + REG_SD_RSP0);
			DBG_PRT_GEN("resp    : 0x%08x\n", cmd->response[0]);
		}
	}

	if(data){
		if(dma){
			mmc_dma_status();
			mmc_dma_set_mode(host, data->flags);
			if(data->flags == MMC_DATA_READ){
				mmc_dma_set_transfer(host, data->flags, (u32)data->dest, (data->blocks * data->blocksize));
				vaddr_purge_cache((unsigned long)data->dest,
					data->blocks * data->blocksize,
					PURGE_CACHE_D_INV_BEFORE_DMA /*PURGE_CACHE_D_PURGE_INV*/);
			}else{
				mmc_dma_set_transfer(host, data->flags, (u32)data->src,  (data->blocks * data->blocksize));
				vaddr_purge_cache((unsigned long)data->src,
					data->blocks * data->blocksize,
					PURGE_CACHE_D_PURGE_INV); /* PURGE_CACHE_D_PURGE */
			}
			mmc_dma_start(host, data->flags);
			
			if (mmc_dma_checkdmaend(host, data->flags)) {
				printf("mmc_dma_checkdmaend error\n");
				ret = 1;
				goto err;
			}
			if (data->flags == MMC_DATA_READ) {
				vaddr_purge_cache((unsigned long)data->dest,
					data->blocks * data->blocksize,
					PURGE_CACHE_D_INV_AFTER_DMA /*PURGE_CACHE_D_PURGE_INV*/);
			}
			/* eMMC buffer DMA transfer set(disable) */
			maskwrite_32x1(sdip_base + REG_CC_EXT_MODE, 0x0002, 0x0000);
		}
		else{
			if(data->flags == MMC_DATA_READ){
				if(uniphier_mmc_read_buf(mmc, (data->blocks * data->blocksize), (u16 *)data->dest)){
					printf("uniphier_mmc_read_buf error\n");
					ret = 1;
					goto err;
				}
			}else{
				if(uniphier_mmc_write_buf(mmc, (data->blocks * data->blocksize), (u16 *)data->src)){
					printf("uniphier_mmc_write_buf error\n");
					ret = 1;
					goto err;
				}
			}
			/* Wait All accesses complete */
			if(uniphier_mmc_data_end(mmc)){
				printf("uniphier_mmc_data_end error\n");
				ret = 1;
				goto err;
			}
		}
	}

err:
	return ret;
}

static void uniphier_set_ios(struct mmc *mmc)
{
	uniphier_mmc_t *host = (uniphier_mmc_t *)mmc->priv;
	u32 sdip_base = host->reg_sdip_base;

	uint div_ratio, div, clk_ctrl;

	DBG_PRT_FUNC("\n");

	/* Set bus width */
	if (mmc->bus_width == 8) 
		maskwrite_32x1(sdip_base + REG_SD_OPTION, 0x0000a000, 0x00006000);
	else if (mmc->bus_width == 4)
		maskwrite_32x1(sdip_base + REG_SD_OPTION, 0x0000a000, 0x00004000);
	else
		maskwrite_32x1(sdip_base + REG_SD_OPTION, 0x0000a000, 0x0000c000);

	if (!mmc->clock)
		return;

	/* Set the clock speed */
	div_ratio = (mmc->f_max + (mmc->clock - 1)) / mmc->clock;

	if (div_ratio <= 2)
		div = 0x00;
	else if (div_ratio <= 4)
		div = 0x01;
	else if (div_ratio <= 8)
		div = 0x02;
	else if (div_ratio <= 16)
		div = 0x04;
	else if (div_ratio <= 32)
		div = 0x08;
	else if (div_ratio <= 64)
		div = 0x10;
	else if (div_ratio <= 128)
		div = 0x20;
	else if (div_ratio <= 256)
		div = 0x40;
	else if (div_ratio <= 512)
		div = 0x80;
	else
		div = 0x80;

	if (div_ratio <= 1) {
		clk_ctrl = 0x00000500;
	} else {
		clk_ctrl = 0x00000100 | (div & 0xff);
	}

	DBG_PRT_GEN("div_ratio %d div %d set_clk %d max_clk %d\n",
		div_ratio, div, mmc->clock, mmc->f_max);

	write_32x1(sdip_base + REG_SD_CLK_CTRL, clk_ctrl);
}

static int uniphier_send_cmd(struct mmc *mmc, struct mmc_cmd *cmd, struct mmc_data *data)
{
	int ret = 0;
	int status;
	
	DBG_PRT_FUNC("\n");

	status = uniphier_send_command_main(mmc, cmd, data);
	if (status & 0x0040) {
		DBG_PRT_GEN("time out error!\n");
		return TIMEOUT;
	}

	return ret;
}

/* Entered into mmc structure during driver init */
static int uniphier_init(struct mmc *mmc)
{
	uniphier_mmc_t *host = (uniphier_mmc_t *)mmc->priv;
	u32 sdip_base = host->reg_sdip_base;
#if defined(CONFIG_MN2WS0220_ref) || defined(CONFIG_MN2WS0230_ref)
	int id = host->id;
#endif /* CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref) */
	
	DBG_PRT_FUNC("\n");

	/* Clock stop */
	maskwrite_32x1(sdip_base + REG_SD_CLK_CTRL, 0x00000100, 0x00000000);

#if defined(CONFIG_MN2WS0220_ref) || defined(CONFIG_MN2WS0230_ref)
	/* Enable clock for MIO DMAC block */
	maskwrite_32x1(SYS_MIO_CLKCTRL(0), 0x02000000, 0x02000000);
#endif /* CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref) */

#if defined(CONFIG_MN2WS0220_ref) || defined(CONFIG_MN2WS0230_ref)
	/* Enable clock for SD I/F ChX block */
	maskwrite_32x1(SYS_MIO_CLKCTRL(id), 0x00000100, 0x00000100);
#else /* CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref) */
	/* Enable clock for SD I/F */
	if (host->enable_sdio == ENABLE_MMC) {
		maskwrite_32x1(SD_CLKCTRL0, SD_CLKCTRL0_SDIO0ON,
			       SD_CLKCTRL0_SDIO0ON_ON);
	}
	else { /* eMMC */
		maskwrite_32x1(SD_CLKCTRL1, SD_CLKCTRL1_SDIO1ON,
			       SD_CLKCTRL1_SDIO1ON_ON);
	}
#endif /* CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref) */

#if defined(CONFIG_MN2WS0220_ref) || defined(CONFIG_MN2WS0230_ref)
	/* Reset SD I/F ChX block */
	maskwrite_32x1(SYS_MIO_RSTCTRL(id), 0x00000001, 0x00000001);
	maskwrite_32x1(SYS_MIO_RSTCTRL(id), 0x00000001, 0x00000000);
	maskwrite_32x1(SYS_MIO_RSTCTRL(id), 0x00000001, 0x00000001);

	/* Reset SD I/F ChX Bridge */
	maskwrite_32x1(SYS_MIO_RSTCTRL(id), 0x04000000, 0x00000000);
	maskwrite_32x1(SYS_MIO_RSTCTRL(id), 0x04000000, 0x04000000);
#else /* CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref) */
	/* Reset SD I/F */
	if (host->enable_sdio == ENABLE_MMC) {
		maskwrite_32x1(SD_RSTCTRL0, SD_RSTCTRL0_SDIO0RST,
			       SD_RSTCTRL0_SDIO0RST_CLEAR);
		maskwrite_32x1(SD_RSTCTRL0, SD_RSTCTRL0_SDIO0RST,
			       SD_RSTCTRL0_SDIO0RST_RESET);
		maskwrite_32x1(SD_RSTCTRL0, SD_RSTCTRL0_SDIO0RST,
			       SD_RSTCTRL0_SDIO0RST_CLEAR);
	}
	else { /* eMMC */
		maskwrite_32x1(SD_RSTCTRL1, SD_RSTCTRL1_SDIO1RST,
			       SD_RSTCTRL1_SDIO1RST_CLEAR);
		maskwrite_32x1(SD_RSTCTRL1, SD_RSTCTRL1_SDIO1RST,
			       SD_RSTCTRL1_SDIO1RST_RESET);
		maskwrite_32x1(SD_RSTCTRL1, SD_RSTCTRL1_SDIO1RST,
			       SD_RSTCTRL1_SDIO1RST_CLEAR);
	}
#endif /* CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref) */

#if 0
	if (host->enable_sdio == ENABLE_EMMC) {
		maskwrite_32x1(SYS_MIO_EMMC0_HWRST(id), 0x00000001, 0x00000001);
		maskwrite_32x1(SYS_MIO_EMMC0_HWRST(id), 0x00000001, 0x00000000);
		maskwrite_32x1(SYS_MIO_EMMC0_HWRST(id), 0x00000001, 0x00000001);

	}
#endif
	

#if !(defined(CONFIG_MN2WS0220_ref) || defined(CONFIG_MN2WS0230_ref))
	/* Enable 32bit access */
	maskwrite_32x1(sdip_base + REG_HOST_MODE, 0x00000101, 0x00000101);
#endif /* !(CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref) */

#if defined(CONFIG_MN2WS0220_ref) || defined(CONFIG_MN2WS0230_ref)
	/* IMCLK 100MHz */
	maskwrite_32x1(SYS_MIO_MODECTRL(id), 0x00071300, 0x00041000);

	/* Set MMC Mode */
	maskwrite_32x1(SYS_MIO_MODECTRL(id), 0x08000000,
		(host->enable_sdio == ENABLE_EMMC) ? 0x08000000 : 0x00000000);

	DBG_PRT_GEN("MIO_CLKCTRL%d 0x%08x\n", id, read_32x1(SYS_MIO_CLKCTRL(id)));
	DBG_PRT_GEN("MIO_RSTCTRL%d 0x%08x\n", id, read_32x1(SYS_MIO_RSTCTRL(id)));
	DBG_PRT_GEN("MIO_MODECTRL%d 0x%08x\n", id, read_32x1(SYS_MIO_MODECTRL(id)));
#else /* CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref) */
	if (host->enable_sdio == ENABLE_MMC) {
		maskwrite_32x1(SD_MODECTRL0,
			       SD_MODECTRL0_mmcmode0 |
			       SD_MODECTRL0_BswapenSD0 |
			       SD_MODECTRL0_sdclksel0_sdclkmod0,
			       SD_MODECTRL0_mmcmode0_SD |
			       SD_MODECTRL0_BswapenSD0_NOSWAP |
			       SD_MODECTRL0_sdclksel0_sdclkmod0_100MHZ
			       );
	}
	else { /* eMMC */
		maskwrite_32x1(SD_MODECTRL1,
			       SD_MODECTRL1_mmcmode1 |
			       SD_MODECTRL1_BswapenSD1 |
			       SD_MODECTRL1_sdclksel1_sdclkmod1,
			       SD_MODECTRL1_mmcmode1_EMMC |
			       SD_MODECTRL1_BswapenSD1_NOSWAP |
			       SD_MODECTRL1_sdclksel1_sdclkmod1_100MHZ
			       );
	}
#endif /* CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref) */

	/* Soft reset */
	maskwrite_32x1(sdip_base + REG_SOFT_RST, 0x00000001, 0x00000001);

	/* Set non-DDR mode */
	maskwrite_32x1(sdip_base + REG_SDIF_MODE, 0x00000001, 0x00000000);

	/* ALL irq mask */
	write_32x1(sdip_base + REG_SD_INFO1_MASK, SD_INFO1_MASK_ALL);
	write_32x1(sdip_base + REG_SD_INFO2_MASK, SD_INFO2_MASK_ALL);
	maskwrite_32x1(sdip_base + REG_SD_INFO1_MASK, SD_INFO1_INFO_CARD_ALL,
		0x00000000);
	
	return 0;
}

static int uniphier_initialize(bd_t *bis, int sdif_id, int sdio)
{
	struct mmc *mmc;
	struct uniphier_mmc *host;
	
	DBG_PRT_FUNC("\n");

	if (sdif_id < 0 || sdif_id > 1)
		return -ENODEV;

	if (!(sdio == ENABLE_MMC || sdio == ENABLE_EMMC))
		return -ENODEV;

	mmc = malloc(sizeof(struct mmc));

	if (!mmc)
		return -ENOMEM;

	memset(mmc, 0, sizeof(struct mmc));

	host = malloc(sizeof(struct uniphier_mmc));
	if (!host) {
		free(mmc);
		return -ENOMEM;
	}

	host->id = sdif_id;
	host->enable_sdio = sdio;
	host->appcmd_flg = 0;

#if defined(CONFIG_MN2WS0220_ref) || defined(CONFIG_MN2WS0230_ref)
	switch (sdif_id) {
	case 0:
		host->reg_sdip_base = 0x5a400000;
		host->dma_ch = 4;
		break;

	case 1:
		host->reg_sdip_base = 0x5a500000;
		host->dma_ch = 5;
		break;

	default:
		break;
	}
#endif /* CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref */

	switch (sdio) {
	case ENABLE_MMC:
		strcpy(mmc->name, "mn_mmc");
		mmc->host_caps = MMC_MODE_4BIT | MMC_MODE_HS | MMC_MODE_HC;
#if !(defined(CONFIG_MN2WS0220_ref) || defined(CONFIG_MN2WS0230_ref))
		host->reg_sdip_base = MMC_IF_BASEADDR;
#endif /* !(CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref) */
		break;

	case ENABLE_EMMC:
		strcpy(mmc->name, "mn_emmc");
		mmc->host_caps = MMC_MODE_4BIT |
				 MMC_MODE_8BIT |
				 MMC_MODE_HS | MMC_MODE_HS_52MHz | MMC_MODE_HC;
#if !(defined(CONFIG_MN2WS0220_ref) || defined(CONFIG_MN2WS0230_ref))
		host->reg_sdip_base = EMMC_IF_BASEADDR;
#endif /* !(CONFIG_MN2WS0220_ref || CONFIG_MN2WS0230_ref) */
		break;

	default:
		break;
	}

	mmc->priv = host;
	
	/* functions */
	mmc->send_cmd = uniphier_send_cmd;
	mmc->set_ios = uniphier_set_ios;
	mmc->init = uniphier_init;

	mmc->voltages = MMC_VDD_32_33 | MMC_VDD_33_34;	/* 3.3V */

	mmc->f_max = 100000000;
	mmc->f_min = (mmc->f_max + 511) / 512;

	mmc->b_max = 0x7fff;

	mmc_register(mmc);

	return 0;
}

int uniphier_mmc_init(bd_t *bis)
{
	int ret = -1;

#ifdef CONFIG_ENABLE_EMMC_CH0
	if (uniphier_initialize(bis, 0, ENABLE_EMMC) >= 0)
		ret = 1;
#endif

#ifdef CONFIG_ENABLE_EMMC_CH1
	if (uniphier_initialize(bis, 1, ENABLE_EMMC) >= 0)
		ret = 1;
#endif

#ifdef CONFIG_ENABLE_MMC_CH0
	if (uniphier_initialize(bis, 0, ENABLE_MMC) >= 0)
		ret = 1;
#endif

#ifdef CONFIG_ENABLE_MMC_CH1
	if (uniphier_initialize(bis, 1, ENABLE_MMC) >= 0)
		ret = 1;
#endif

	return ret;
}
